﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text.RegularExpressions;

using LibGit2Sharp;

using NAnt.Core;
using NAnt.Core.Attributes;

namespace Nant.Git
{
    [TaskName( "git.fetch-version.number-and-revision.hash-properties" )]
    public class GetVersionNumberBasedOnTags : Task
    {
        public const string RepositoryNameProperty = "repo.dir";
        public const string MajorMinorVersionProperty = "base.version";
        public const string VersionNumberProperty = "version.number";
        public const string RevisionHashProperty = "revision.hash";

        #region Overrides of Task

        public override bool Verbose
        {
            get
            {
                return true;
            }
            set
            {
            }
        }

        protected override void ExecuteTask()
        {
            if ( Properties[ RepositoryNameProperty ] == null )
            {
                throw new BuildException( string.Format( "Property '{0}' is not provided.", RepositoryNameProperty ) );
            }
            if ( Properties[ MajorMinorVersionProperty ] == null )
            {
                throw new BuildException( string.Format( "Property '{0}' is not provided.", MajorMinorVersionProperty ) );
            }

            string repositoryLocation = Properties[ RepositoryNameProperty ];

            repositoryLocation = Path.GetFullPath( Environment.ExpandEnvironmentVariables( repositoryLocation ) );

            string baseVersion = Properties[ MajorMinorVersionProperty ];

            MatchCollection versionDigits = new Regex( "(?'verinfo'[0-9]+)" ).Matches( baseVersion );

            List<string> digits = new List<string>();

            foreach ( Match versionDigit in versionDigits )
            {
                digits.Add( versionDigit.Value );
            }

            while ( digits.Count > 3 )
            {
                digits.Remove( digits.Last() );
            }

            while ( digits.Count < 3 )
            {
                digits.Add( "0" );
            }

            string version = string.Format( "{0}.{1}.{2}.", digits[ 0 ], digits[ 1 ], digits[ 2 ] );

            Repository repo;

            try
            {
                repo = new Repository( repositoryLocation );
            }
            catch ( Exception ex )
            {
                throw new BuildException( string.Format( "Repository at '{0}' is invalid.", repositoryLocation ), ex );
            }

            Tag tag = repo.Tags.LastOrDefault( t => t.Target is Commit );

            if ( tag == null )
            {
                throw new BuildException( "There are no tags in repository." );
            }

            Commit lastCommit = repo.Commits.FirstOrDefault();

            if ( lastCommit == null )
            {
                throw new BuildException( "There are no commits in repository." );
            }

            Commit tagCommit = ( Commit )tag.Target;

            int numberOfCommitsSinceLastTag = 0;

            foreach ( Commit commit in repo.Commits )
            {
                ++numberOfCommitsSinceLastTag;
                if ( commit.Sha == tagCommit.Sha )
                {
                    break;
                }
            }

            if ( Properties.Contains( RevisionHashProperty ) )
            {
                Properties.Remove( RevisionHashProperty );
            }

            if ( Properties.Contains( VersionNumberProperty ) )
            {
                Properties.Remove( VersionNumberProperty );
            }

            Properties.Add( RevisionHashProperty, lastCommit.Sha );
            Properties.Add( VersionNumberProperty, version + numberOfCommitsSinceLastTag );
        }

        #endregion
    }
}