﻿using System;
using System.Data.SqlClient;
using System.Diagnostics;
using System.IO;
using System.Text;
using NAnt.Core;
using NAnt.Core.Attributes;

namespace Nant.SQLServer
{
    [TaskName("exec-sqlserver-sql")]
    public class RunScript : Task
    {
        public const string ConnectionStringPropertyName = "sqlserver.connection.string";

        [TaskAttribute("file", Required = false)]
        public string ScriptFilePath { get; set; }

        [TaskAttribute("connection-string", Required = false)]
        public string ConnectionString { get; set; }

        [TaskAttribute("transaction", Required = false)]
        [BooleanValidator]
        public bool IsTransacted { get; set; }

        protected override void ExecuteTask()
        {
            string connectionString = GetConnectionString();

            SqlConnection connection = GetValidatedConnection(connectionString);

            using (MemoryStream stream = new MemoryStream())
            {
                StreamWriter writer = new StreamWriter(stream);

                writer.Write(Project.ExpandProperties(XmlNode.InnerText, Location));
                writer.Flush();
                stream.Position = 0;

                ExecuteSql(stream, connection);
            }

            string scriptFilePath = !string.IsNullOrEmpty(ScriptFilePath) ? Path.GetFullPath(Environment.ExpandEnvironmentVariables(ScriptFilePath)) : null;

            if (File.Exists(scriptFilePath))
            {
                Log(Level.Info, "Running script file: " + scriptFilePath);
                using (FileStream stream = new FileStream(scriptFilePath, FileMode.Open, FileAccess.Read))
                    ExecuteSql(stream, connection);
            }
        }

        private void ExecuteSql(Stream stream, SqlConnection connection)
        {
            StreamReader reader = new StreamReader(stream, Encoding.UTF8);

            //TODO: add batch running when script is extremely large

            string[] items = reader.ReadToEnd().Split(new[] {"GO"}, StringSplitOptions.RemoveEmptyEntries);
            reader.Close();

            SqlCommand command = new SqlCommand
                {
                    Connection = connection
                };
            SqlTransaction transaction = null;

            connection.Open();

            if (IsTransacted)
                transaction = connection.BeginTransaction();
            command.Transaction = transaction;

            string lastItem = null;

            try
            {
                foreach (string item in items)
                {
                    command.CommandText = lastItem = item;
                    Log(Level.Info, "Affected row: " + command.ExecuteNonQuery());
                }
                
                if (transaction != null) transaction.Commit();
            }
            catch
            {
                Log(Level.Verbose, lastItem);
                if (transaction != null) transaction.Rollback();
                throw;
            }
            finally
            {
                connection.Close();
            }
        }

        private SqlConnection GetValidatedConnection(string connectionString)
        {
            SqlConnection connection = new SqlConnection(connectionString);

            try
            {
                connection.Open();
                connection.Close();
            }
            catch (Exception ex)
            {
                throw new BuildException("Failed to connect to a server using provided connection string: " + connectionString, ex);
            }

            return connection;
        }


        private string GetConnectionString()
        {
            if (string.IsNullOrEmpty(Properties[ConnectionStringPropertyName]) && string.IsNullOrEmpty(ConnectionString))
                throw new BuildException("Please, provide a connection string either via 'connection-string' attribute or '" + ConnectionStringPropertyName + "' property.");

            return string.IsNullOrEmpty(Properties[ConnectionStringPropertyName]) ? ConnectionString : Properties[ConnectionStringPropertyName];
        }
    }
}