﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Net;
using System.Text.RegularExpressions;

using Microsoft.TeamFoundation.Client;
using Microsoft.TeamFoundation.VersionControl.Client;

using NAnt.Core;
using NAnt.Core.Attributes;

namespace Nant.Tfs
{

    [TaskName( "tfs.fetch-version.number-and-revision.hash-properties" )]
    public class TfsVersionNumberBasedOnChangesets : Task
    {
        public const string MajorMinorVersionProperty = "base.version";

        public const string VersionNumberProperty = "version.number";
        public const string RevisionHashProperty = "revision.hash";

        public const string TfsProjectPathPropertyName = "tfs.project.path";

        public const string TfsUrlPropertyName = "tfs.server.url";
        public const string TfsUserNamePropertyName = "tfs.user.name";
        public const string TfsUserPasswordPropertyName = "tfs.user.password";

        #region Overrides of Task

        protected override void ExecuteTask()
        {
            if ( Properties[ MajorMinorVersionProperty ] == null )
            {
                throw new BuildException( string.Format( "Property '{0}' is not provided.", MajorMinorVersionProperty ) );
            }
            if ( Properties[ TfsUrlPropertyName ] == null )
            {
                throw new BuildException( string.Format( "Property '{0}' is not provided.", TfsUrlPropertyName ) );
            }
            if ( Properties[ TfsUserNamePropertyName ] == null )
            {
                throw new BuildException( string.Format( "Property '{0}' is not provided.", TfsUserNamePropertyName ) );
            }
            if ( Properties[ TfsUserPasswordPropertyName ] == null )
            {
                throw new BuildException( string.Format( "Property '{0}' is not provided.", TfsUserPasswordPropertyName ) );
            }
            if ( Properties[ TfsProjectPathPropertyName ] == null )
            {
                throw new BuildException( string.Format( "Property '{0}' is not provided.", TfsProjectPathPropertyName ) );
            }

            string baseVersion = Properties[ MajorMinorVersionProperty ];

            MatchCollection versionDigits = new Regex( "(?'verinfo'[0-9]+)" ).Matches( baseVersion );

            List<string> digits = new List<string>();

            foreach ( Match versionDigit in versionDigits )
            {
                digits.Add( versionDigit.Value );
            }

            while ( digits.Count > 3 )
            {
                digits.Remove( digits.Last() );
            }

            while ( digits.Count < 3 )
            {
                digits.Add( "0" );
            }

            string version = string.Format( "{0}.{1}.{2}.", digits[ 0 ], digits[ 1 ], digits[ 2 ] );
            string revision = "0";

            TfsTeamProjectCollection tfs = new TfsTeamProjectCollection(
                new Uri( Properties[ TfsUrlPropertyName ] ),
                new NetworkCredential( Properties[ TfsUserNamePropertyName ], Properties[ TfsUserPasswordPropertyName ] ) );

            VersionControlServer versionControl = tfs.GetService<VersionControlServer>();

            VersionSpec versionFrom = null;
            VersionSpec versionTo = VersionSpec.Latest;

            IEnumerable<Changeset> changesets = versionControl.QueryHistory(
                Properties[ TfsProjectPathPropertyName ],
                VersionSpec.Latest,
                0,
                RecursionType.Full,
                null,
                versionFrom,
                versionTo,
                Int32.MaxValue,
                true,
                false,
                false,
                true
                ).Cast<Changeset>();

            Changeset lastChangeset = changesets.LastOrDefault();

            if ( lastChangeset == null )
            {
                version += 0;
            }
            else
            {
                version += changesets.Count();
                revision = lastChangeset.ChangesetId.ToString( CultureInfo.InvariantCulture );
            }

            if ( Properties.Contains( RevisionHashProperty ) )
            {
                Properties.Remove( RevisionHashProperty );
            }

            if ( Properties.Contains( VersionNumberProperty ) )
            {
                Properties.Remove( VersionNumberProperty );
            }

            Properties.Add( RevisionHashProperty, revision );
            Properties.Add( VersionNumberProperty, version );
        }

        #endregion
    }
}