﻿using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.IO;
using System.Linq;
using System.Net;
using System.Threading;

using VMware.Vim;

using FileInfo = System.IO.FileInfo;

namespace Nant.VMWare
{
    public class TestEnvironment : IDisposable
    {
        private readonly NamePasswordAuthentication _authentication;
        private readonly GuestOperationsManager _manager;
        private readonly VimClient _virtualHost;
        private readonly VirtualMachine _vm;
        private readonly VmWareConfiguration _vmWareConfiguration;
        private GuestFileManager _fileManager;

        public TestEnvironment( VmWareConfiguration vmWareConfiguration )
        {
            _vmWareConfiguration = vmWareConfiguration;

            _virtualHost = new VimClient();
            _virtualHost.Login(
                string.Format( "https://{0}/sdk", _vmWareConfiguration.FarmUrl ),
                _vmWareConfiguration.FarmLogin,
                _vmWareConfiguration.FarmPassword );

            _vm = ( VirtualMachine )_virtualHost.FindEntityView(
                typeof( VirtualMachine ),
                null,
                new NameValueCollection
                    {
                        {"name", _vmWareConfiguration.VirtualMachineName}
                    },
                null );

            _manager = new GuestOperationsManager( _virtualHost, _virtualHost.ServiceContent.GuestOperationsManager );

            _authentication = new NamePasswordAuthentication
                                  {
                                      Username = _vmWareConfiguration.User,
                                      Password = _vmWareConfiguration.Password,
                                      InteractiveSession = true
                                  };
        }

        #region IDisposable Members

        public void Dispose()
        {
            _vm.ShutdownGuest();
            _virtualHost.Disconnect();
        }

        #endregion

        public void CopyFiles( string sourceFolder, string destinationFolderInVm )
        {
            if ( !sourceFolder.EndsWith( "\\" ) )
            {
                sourceFolder = sourceFolder + "\\";
            }

            string[] fileToCopy = Directory.GetFiles( sourceFolder, "*.*", SearchOption.AllDirectories );

            Console.WriteLine( "copy file: " + fileToCopy.Length );

            foreach ( string file in fileToCopy )
            {
                string fileName = file.Replace( sourceFolder, "" );
                string targetFileFullPath = Path.Combine( destinationFolderInVm, fileName );

                Console.Write( file );
                Console.Write( " -> " );
                Console.WriteLine( targetFileFullPath );

                CopyFile( file, targetFileFullPath );
            }
        }

        public void AuthenticateInGuest()
        {
            _manager.UpdateViewData();

            GuestAuthManager auth = new GuestAuthManager( _virtualHost, _manager.AuthManager );

            while ( true )
            {
                try
                {
                    auth.ValidateCredentialsInGuest( _vm.MoRef, _authentication );
                    break;
                }
                catch ( VimException )
                {
                    Thread.Sleep( 10000 );
                }
            }

            _fileManager = new GuestFileManager( _virtualHost, _manager.FileManager );
        }

        public void InitializeCleanEnvironment()
        {
            VirtualMachineSnapshot cleanSnapshot = FindSnapshot(
                _vmWareConfiguration.CleanSnapshotName, _vm.Snapshot.RootSnapshotList );

            if ( cleanSnapshot == null )
            {
                throw new ApplicationException( "Unable to find a snapshot named " + _vmWareConfiguration.CleanSnapshotName );
            }

            do
            {
                _vm.UpdateViewData();
                Thread.Sleep( 10000 );
            }
            while ( _vm.Runtime.PowerState == VirtualMachinePowerState.poweredOn );

            cleanSnapshot.RevertToSnapshot( null, false );
        }

        private VirtualMachineSnapshot FindSnapshot( string name, IEnumerable<VirtualMachineSnapshotTree> trees )
        {
            VirtualMachineSnapshot snapshot = null;

            foreach ( VirtualMachineSnapshotTree tree in trees )
            {
                if ( tree.Name == name )
                {
                    snapshot = new VirtualMachineSnapshot( _virtualHost, tree.Snapshot );
                    break;
                }

                snapshot = FindSnapshot( name, tree.ChildSnapshotList );

                if ( snapshot != null )
                {
                    break;
                }
            }

            return snapshot;
        }

        public void Start()
        {
            _vm.PowerOnVM( null );

            while ( !_vm.Guest.GuestOperationsReady.HasValue || !_vm.Guest.GuestOperationsReady.Value )
            {
                _vm.UpdateViewData();
                Thread.Sleep( 5000 );
            }

            AuthenticateInGuest();
        }

        public void UploadArtifact( string sourceFilePath, string destinationPath )
        {
            string serviceUrl = _virtualHost.VimService.Url;

            serviceUrl = serviceUrl.Substring( 0, serviceUrl.LastIndexOf( "sdk" ) - 1 );

            string localFilePath = Path.GetFullPath( sourceFilePath );

            String httpUrl = serviceUrl + "/folder/" + destinationPath + "?dcPath=ha-datacenter&dsName=" + _vmWareConfiguration.DataStoreName;

            httpUrl = httpUrl.Replace( " ", " % 20" );

            string cookie = _virtualHost.VimService.CookieContainer.GetCookies( new Uri( _virtualHost.VimService.Url ) )[ 0 ].ToString();

            HttpWebRequest request = ( HttpWebRequest )WebRequest.Create( httpUrl );
            request.Method = "PUT";
            request.Credentials = new NetworkCredential {UserName = _vmWareConfiguration.FarmLogin, Password = _vmWareConfiguration.FarmPassword};
            request.Headers.Add( HttpRequestHeader.Cookie, cookie );
            request.Timeout = Timeout.Infinite;
            request.ReadWriteTimeout = Timeout.Infinite;
            request.ContentType = "application/octet-stream";
            request.AllowWriteStreamBuffering = false;

            const int bytes = 1000;
            byte[] buffer = new byte[bytes];

            using ( FileStream fs = new FileStream( localFilePath, FileMode.Open, FileAccess.Read ) )
            {
                request.ContentLength = fs.Length;

                using ( Stream destination = request.GetRequestStream() )
                {
                    int read;
                    while ( ( read = fs.Read( buffer, 0, bytes ) ) != 0 )
                    {
                        destination.Write( buffer, 0, read );
                    }

                    destination.Close();
                }
                fs.Close();
            }
        }

        public void DeployAndConnectIso( string localPath, string remotePath )
        {
            UploadArtifact( localPath, remotePath );

            _vm.UpdateViewData();

            VirtualCdrom cd = _vm.Config.Hardware.Device.FirstOrDefault( d => d is VirtualCdrom ) as VirtualCdrom;

            if ( cd == null )
            {
                throw new ApplicationException( "CD is null" );
            }

            VirtualCdromIsoBackingInfo newBacking = new VirtualCdromIsoBackingInfo
                                                        {
                                                            FileName = string.Format( "[{0}] {1}", _vmWareConfiguration.DataStoreName, remotePath )
                                                        };
            cd.Backing = newBacking;
            cd.Connectable.StartConnected = true;
            cd.Connectable.Connected = true;

            VirtualDeviceConfigSpec deviceSpec = new VirtualDeviceConfigSpec
                                                     {
                                                         Device = cd,
                                                         Operation = VirtualDeviceConfigSpecOperation.edit
                                                     };

            VirtualMachineConfigSpec vmSpec = new VirtualMachineConfigSpec
                                                  {
                                                      DeviceChange = new[] {deviceSpec}
                                                  };

            _vm.ReconfigVM( vmSpec );
        }

        public int RunProcess( string executable, string args, bool waitForExit = false )
        {
            GuestProcessManager processManager = new GuestProcessManager( _virtualHost, _manager.ProcessManager );

            GuestProgramSpec arguments = new GuestProgramSpec
                                             {
                                                 Arguments = args,
                                                 ProgramPath = executable,
                                                 WorkingDirectory = Path.GetDirectoryName( executable )
                                             };

            long pid = processManager.StartProgramInGuest( _vm.MoRef, _authentication, arguments );
            int exitCode = 0;

            if ( waitForExit )
            {
                while ( true )
                {
                    GuestProcessInfo process = processManager.ListProcessesInGuest( _vm.MoRef, _authentication, new[] {pid} )
                        .FirstOrDefault( p => p.Pid == pid );

                    if ( process == null || process.EndTime != null )
                    {
                        if ( process != null && process.ExitCode.HasValue )
                        {
                            exitCode = process.ExitCode.Value;
                        }
                        break;
                    }

                    Thread.Sleep( 60000 );
                    processManager.UpdateViewData();
                }
            }

            return exitCode;
        }

        public void TakeSnapshot( string name, string description )
        {
            _vm.CreateSnapshot( name, description, true, true );
        }

        public void Shutdown()
        {
            _vm.ShutdownGuest();
        }

        public void CopyFile( string sourceFilePathOnHost, string targetFilePathInEnvironment )
        {
            string folder = Path.GetDirectoryName( targetFilePathInEnvironment );

            bool exists = false;

            try
            {
                GuestListFileInfo info = _fileManager.ListFilesInGuest( _vm.MoRef, _authentication, Path.GetDirectoryName( folder ), null, null,
                                                                        Path.GetFileName( folder ) );

                exists = ( info.Files != null && info.Files.Length > 0 && info.Files.Any( f => f.Path == Path.GetFileName( folder ) ) );
            }
            catch ( VimException )
            {
            }

            try
            {
                if ( !exists )
                {
                    _fileManager.MakeDirectoryInGuest( _vm.MoRef, _authentication, folder, true );
                }
            }
            catch ( VimException )
            {
            }

            string url = _fileManager.InitiateFileTransferToGuest(
                _vm.MoRef,
                _authentication,
                targetFilePathInEnvironment,
                new GuestWindowsFileAttributes(),
                new FileInfo( sourceFilePathOnHost ).Length,
                true );

            UploadFile( sourceFilePathOnHost, url.Replace( "*", _vmWareConfiguration.FarmUrl ) );
        }

        public void GetFile( string sourceFilePathInEnvironment, string targetFilePathOnHost )
        {
            Thread.Sleep( 5000 );

            FileTransferInformation info = _fileManager.InitiateFileTransferFromGuest( _vm.MoRef, _authentication, sourceFilePathInEnvironment );

            DownloadFile( info.Url.Replace( "*", _vmWareConfiguration.FarmUrl ), targetFilePathOnHost );
        }

        private void DownloadFile( string from, string to )
        {
            string cookie = _virtualHost.VimService.CookieContainer.GetCookies( new Uri( _virtualHost.VimService.Url ) )[ 0 ].ToString();

            WebClient webCLient = new WebClient
                                      {
                                          Credentials = new NetworkCredential( _vmWareConfiguration.FarmLogin, _vmWareConfiguration.FarmPassword )
                                      };
            webCLient.Headers.Add( HttpRequestHeader.Cookie, cookie );

            while ( true )
            {
                try
                {
                    webCLient.DownloadFile( from, to );
                    break;
                }
                catch ( Exception ex )
                {
                    Console.WriteLine( ex.Message );

                    if ( ex.InnerException != null )
                    {
                        Console.WriteLine( ex.InnerException.Message );
                    }

                    continue;
                }
            }
            webCLient.Dispose();
        }

        private void UploadFile( string from, string to )
        {
            string cookie = _virtualHost.VimService.CookieContainer.GetCookies( new Uri( _virtualHost.VimService.Url ) )[ 0 ].ToString();

            WebClient wc = new WebClient
                               {
                                   Credentials = new NetworkCredential( _vmWareConfiguration.FarmLogin, _vmWareConfiguration.FarmPassword )
                               };
            wc.Headers.Add( HttpRequestHeader.Cookie, cookie );

            while ( true )
            {
                try
                {
                    wc.UploadFile( to, "PUT", from );
                    break;
                }
                catch ( Exception ex )
                {
                    Console.WriteLine( ex.Message );

                    if ( ex.InnerException != null )
                    {
                        Console.WriteLine( ex.InnerException.Message );
                    }

                    Thread.Sleep( 5000 );
                    continue;
                }
            }

            wc.Dispose();
        }

        public void TakeScreenshot( string destination )
        {
            string url = _vm.CreateScreenshot();

            string requestUrl = string.Format(
                "https://{0}/folder/{1}?dcPath=ha-datacenter&dsName={2}",
                _vmWareConfiguration.FarmUrl, url.Replace( "[" + _vmWareConfiguration.DataStoreName + "] ", "" ), _vmWareConfiguration.DataStoreName );

            DownloadFile( requestUrl, destination );
        }

        public string GetHostFolderPath()
        {
            return Path.GetDirectoryName( new Uri( GetType().Assembly.CodeBase ).LocalPath );
        }

        public bool FileExists( string path )
        {
            GuestListFileInfo info = _fileManager.ListFilesInGuest( _vm.MoRef, _authentication, path, null, null, "*.*" );

            return info.Files != null && info.Files.Length > 0;
        }

        public string ReadFile( string path )
        {
            string tempFile = Path.GetTempFileName();

            string contents;

            try
            {
                FileTransferInformation info = _fileManager.InitiateFileTransferFromGuest( _vm.MoRef, _authentication, path );

                DownloadFile( info.Url, tempFile );
                contents = File.ReadAllText( tempFile );
            }
            finally
            {
                File.Delete( tempFile );
            }

            return contents;
        }
    }
}