﻿using System;
using System.Collections.Specialized;
using System.Configuration;
using System.Diagnostics;

using NUnit.Framework;

using VMware.Vim;

namespace Nant.VMWare
{
    [TestFixture]
    [Category( TestCategories.AcceptanceStage )]
    internal partial class SphereVirtualMachineTests
    {
        /*
         * Requirements for vSphere:
         *      - paid license or see a hack
         *      - VI Client installed
         *      - VM's user name has a password
         *      - VM has autologon enabled
         */

        #region Setup/Teardown

        [TearDown]
        public void PowerOffVirtualMachine()
        {
            if ( _virtualMachine != null )
            {
                while ( _virtualMachine.VirtualMachineIsRunning() )
                {
                    try
                    {
                        _vmWareVirtualMachine.ShutdownGuest();
                        break;
                    }
                    catch ( Exception )
                    {
                        _vmWareVirtualMachine.UpdateViewData();
                    }
                }

                Assert.IsFalse( _virtualMachine.VirtualMachineIsRunning() );
            }
        }

        #endregion

        private string FarmUrl
        {
            get
            {
                return ConfigurationManager.AppSettings[ "FarmUrl" ];
            }
        }

        private string FarmLogin
        {
            get
            {
                return ConfigurationManager.AppSettings[ "FarmLogin" ];
            }
        }

        private string FarmPassword
        {
            get
            {
                return ConfigurationManager.AppSettings[ "FarmPassword" ];
            }
        }

        private string VirtualMachineName
        {
            get
            {
                return ConfigurationManager.AppSettings[ "VMName" ];
            }
        }

        private string VirtualMachineUserName
        {
            get
            {
                return ConfigurationManager.AppSettings[ "VMUserName" ];
            }
        }

        private string VirtualMachineUserPassword
        {
            get
            {
                return ConfigurationManager.AppSettings[ "VMUserPassword" ];
            }
        }

        private VirtualMachine _vmWareVirtualMachine;

        [TestFixtureSetUp]
        public void ConnectToFarm()
        {
            VimClient virtualHost = new VimClient();
            virtualHost.Login(
                string.Format( "https://{0}/sdk", FarmUrl ),
                FarmLogin,
                FarmPassword );

            _vmWareVirtualMachine = ( VirtualMachine )virtualHost.FindEntityView(
                typeof( VirtualMachine ),
                null,
                new NameValueCollection
                    {
                        {"name", VirtualMachineName}
                    },
                null );
        }

        [TestFixtureTearDown]
        public void DisconnectFromFarm()
        {
            VimClient client = _vmWareVirtualMachine.Client;

            try
            {
                if ( _vmWareVirtualMachine.Guest.GuestState == "running" )
                {
                    _vmWareVirtualMachine.ShutdownGuest();
                }
                _vmWareVirtualMachine = null;
            }
            catch ( Exception ex )
            {
                Trace.WriteLine( ex.Message );
            }

            client.Logout();
        }

        private IVirtualMachine _virtualMachine;

        private IVirtualMachine CreateVirtualMachine()
        {
            return _virtualMachine = new SphereVirtualMachine( _vmWareVirtualMachine );
        }

        [Test]
        public void DoesNothingWhenVirtualMachineIsOff()
        {
            IVirtualMachine vm = CreateVirtualMachine();
            vm.Shutdown();

            Assert.IsFalse( vm.VirtualMachineIsRunning(), "VM is on." );
        }

        [Test]
        public void LoggedInToVirtualMachineCallReturnsTrueWhenVirtualMachineIsInInteractiveModeAndFalseOtherwise()
        {
            IVirtualMachine vm = CreateVirtualMachine();

            vm.Start();

            Assert.IsFalse( vm.LoggedInToVirtualMachine() );
            vm.Login( VirtualMachineUserName, VirtualMachineUserPassword );
            Assert.IsTrue( vm.LoggedInToVirtualMachine() );
        }

        [Test]
        public void LogsInUserWhenLoginIsCalled()
        {
            IVirtualMachine vm = CreateVirtualMachine();
            vm.Start();

            vm.Login( VirtualMachineUserName, VirtualMachineUserPassword );

            Assert.IsTrue( vm.LoggedInToVirtualMachine(), "User is not logged in to VM" );
        }

        [Test]
        public void RevertsToPoweredOnVmWhenSnapshotWasDoneOnPoweredOnVirtualMachine()
        {
            IVirtualMachine vm = CreateVirtualMachine();
            vm.Start();

            string snapshotName = Guid.NewGuid().ToString( "N" );

            ManagedObjectReference initialSnapshot = _vmWareVirtualMachine.Snapshot.CurrentSnapshot;

            try
            {
                vm.TakeSnapshot( snapshotName, "Some description" );

                VirtualMachineSnapshot snapshot = new VirtualMachineSnapshot( _vmWareVirtualMachine.Client, initialSnapshot );
                snapshot.RevertToSnapshot( null, false );
                _vmWareVirtualMachine.UpdateViewData();

                Assert.IsFalse( vm.VirtualMachineIsRunning() );

                vm.RevertToSnapshot( snapshotName );

                Assert.IsTrue( vm.VirtualMachineIsRunning() );
            }
            finally
            {
                if ( initialSnapshot.ToString() != _vmWareVirtualMachine.Snapshot.CurrentSnapshot.ToString() )
                {
                    VirtualMachineSnapshot snapshot = new VirtualMachineSnapshot( _vmWareVirtualMachine.Client, _vmWareVirtualMachine.Snapshot.CurrentSnapshot );
                    snapshot.UpdateViewData();
                    snapshot.RemoveSnapshot( false, null );
                }
            }
        }

        [Test]
        public void StartsVirtualMachineWhenStartIsCalled()
        {
            IVirtualMachine vm = CreateVirtualMachine();
            vm.Start();

            Assert.IsTrue( vm.VirtualMachineIsRunning(), "VM is off." );
        }

        [Test]
        public void StopsVirtualMachineWhenShutdownIsCalled()
        {
            IVirtualMachine vm = CreateVirtualMachine();
            vm.Start();

            Assert.IsTrue( vm.VirtualMachineIsRunning(), "VM is off." );

            vm.Shutdown();

            Assert.IsFalse( vm.VirtualMachineIsRunning(), "VM is on." );
        }

        [Test]
        public void TakesSnapshotOnPoweredOffVirtualMachineWhenTakeSnapshotIsCalledAndVirtualMachineIsOff()
        {
            IVirtualMachine vm = CreateVirtualMachine();

            string snapshotName = Guid.NewGuid().ToString( "N" );

            ManagedObjectReference initialSnapshot = _vmWareVirtualMachine.Snapshot.CurrentSnapshot;

            try
            {
                vm.TakeSnapshot( snapshotName, "Some description" );
                _vmWareVirtualMachine.UpdateViewData();
            }
            finally
            {
                if ( initialSnapshot.ToString() != _vmWareVirtualMachine.Snapshot.CurrentSnapshot.ToString() )
                {
                    VirtualMachineSnapshot snapshot = new VirtualMachineSnapshot( _vmWareVirtualMachine.Client, _vmWareVirtualMachine.Snapshot.CurrentSnapshot );
                    snapshot.UpdateViewData();
                    snapshot.RemoveSnapshot( false, null );
                }
            }
        }

        [Test]
        public void TakesSnapshotOnPoweredOnVirtualMachineWhenTakeSnapshotIsCalledAndVirtualMachineIsOn()
        {
            IVirtualMachine vm = CreateVirtualMachine();
            vm.Start();

            string snapshotName = Guid.NewGuid().ToString( "N" );

            ManagedObjectReference initialSnapshot = _vmWareVirtualMachine.Snapshot.CurrentSnapshot;

            try
            {
                vm.TakeSnapshot( snapshotName, "Some description" );
                _vmWareVirtualMachine.UpdateViewData();
            }
            finally
            {
                if ( initialSnapshot.ToString() != _vmWareVirtualMachine.Snapshot.CurrentSnapshot.ToString() )
                {
                    VirtualMachineSnapshot snapshot = new VirtualMachineSnapshot( _vmWareVirtualMachine.Client, _vmWareVirtualMachine.Snapshot.CurrentSnapshot );
                    snapshot.UpdateViewData();
                    snapshot.RemoveSnapshot( false, null );
                }
            }
        }

        [Test]
        [ExpectedException( typeof( ArgumentNullException ) )]
        public void ThrowsAnExceptionWhenConstructedWithNullVmWareObject()
        {
            new SphereVirtualMachine( null );
        }

        [Test]
        [ExpectedException( typeof( VirtualMachineIsNotReadyException ) )]
        public void ThrowsAnExceptionWhenLoginIsCalledOnStoppedVirtualMachine()
        {
            IVirtualMachine vm = CreateVirtualMachine();

            vm.Login( "Admin", "Admin" );
        }

        [Test]
        [ExpectedException( typeof( ApplicationException ) )]
        public void ThrowsAnExceptionWhenRevertToSnapshotIsCalledAndSnapshotDoesNotExist()
        {
            IVirtualMachine vm = CreateVirtualMachine();
        }

        [Test]
        public void VirtualMachineIsRunningCallReturnsTrueWhenVirtualMachineIsRunningAndFalseOtherwise()
        {
            IVirtualMachine vm = CreateVirtualMachine();

            Assert.IsFalse( vm.VirtualMachineIsRunning(), "VM must be powered off." );
            vm.Start();
            Assert.IsTrue( vm.VirtualMachineIsRunning(), "VM must be powered on." );
        }
    }
}